import urllib
import json
import julia_semver
from . import _jill
from . import util


_ALL_JULIA_VERSIONS = None
# Pretty heavy, this every release and rc, etc. since the beginning
_JULIA_VERSIONS_URL = 'https://julialang-s3.julialang.org/bin/versions.json'


# From https://github.com/cjdoris/pyjuliapkg
def _all_julia_versions():
    global _ALL_JULIA_VERSIONS
    if _ALL_JULIA_VERSIONS is None:
        url = _JULIA_VERSIONS_URL
        util.log(f'Querying Julia versions from {url}')
        with urllib.request.urlopen(url) as fp:
            _ALL_JULIA_VERSIONS = json.load(fp)
    return _ALL_JULIA_VERSIONS


def _best_version(version_spec=None):
    if version_spec is None:
        version_spec = julia_semver.semver_spec("^1")
    if isinstance(version_spec, str): # else assume it is created by semver_spec
        version_spec = julia_semver.semver_spec(version_spec)
    version_info = _all_julia_versions()
    maxv = julia_semver.version("0.0.0")
    for version in version_info.keys():
        vers = julia_semver.version(version)
        if julia_semver.match(version_spec, vers) and vers > maxv:
            maxv = vers
    if maxv == julia_semver.version("0.0.0"):
        raise Exception(f"No julia version satisfying '{str(version_spec)}' available for download.")
    return maxv


# API of juliaup is not developed enough for this yet
# def install_from_juliaup():

# What should this return ?
def prompt_and_install(
        version_spec=None, answer_yes=False, install_func=None, post_question_hook=None
):
    """
    Download and install julia, optionally prompting, and running a hook after prompting.

    Parameters:
    version_spec - Julia version specification string or object. Default: "^1".
    answer_yes - if `True`, then ask no questions, assume answers are "yes".
    install_func - A function to call that installs julia. The function must accept
        two arguments, `answer_yes`, and `version`. Default : _jill.install_from_jill.
    post_question_hook - a function to run if and after the consumer is asked whether
        to install Julia. This can be used to ask and record more questions rather
        than waiting till after the download. Default: None
    """
    if version_spec is None:
        version_spec = "^1"
    version = str(_best_version(version_spec=version_spec))
    if not answer_yes:
        answer = util.query_yes_no(
            f"Would you like to download and install Julia version '{version}'?"
        )
        if post_question_hook:
            post_question_hook()
        if not answer:
            return False
    if install_func is None:
        install_func = _jill.install_from_jill
    return install_func(answer_yes=answer_yes, version=version)



import os
import shutil
import subprocess
import itertools
import julia_semver
from . import juliaup
from . import _jill
from . import install_julia
from . import util
from . import julia_version


def is_julia_executable(exe):
    """
    Return True if exe is the path to a Julia exectuable, otherwise False.

    The test consists of checking the output of "julia --version".
    """
    try:
        words = subprocess.run(
            [exe, '--version'], check=True, capture_output=True, encoding='utf8'
        ).stdout.strip().split()
    except:
        return False
    return len(words) == 3 and words[0] == "julia" and words[1] == "version"


# Strange to take the argument to skip check. But, it makes call site much simpler
def _check_path(path, check_exe=True):
    if not check_exe:
        return True
    if not os.path.isfile(path):
        return False
    return is_julia_executable(path)


def _to_semver(versions_paths):
    return [(julia_semver.version(v), p) for (v, p) in versions_paths]

def _default_locations(default=True):
    return {'jill': default, 'juliaup': default, 'which': default, 'env': default}

def _collect_paths(locations):
    all_paths = []
    def tolist(x):
        if x is None:
            return []
        return x if isinstance(x, list) else [x]
    # Finding juliaup versions is fast because they are cached in the filesystem.
    if locations['juliaup']:
        paths_juliaup = _to_semver(juliaup.version_path_list())
        all_paths.append(paths_juliaup)
    if locations['which']:
        wpath = shutil.which("julia")
        # Exclude ~/.juliaup/bin/julia . It actually links to julialauncher. This program
        # is not really julia. Eg, if DEPOT_PATH[1] has been changed, julialauncher will error.
        if wpath is not None and (not wpath.find("juliaup") >= 0): # juliaup is not in the path name
            paths_which = _to_semver(julia_version.to_version_path_list(tolist(wpath)))
            all_paths.append(paths_which)
    if locations['jill']:
        paths_jill = _to_semver(_jill.version_path_list())
        all_paths.append(paths_jill)
    return all_paths


def _env_var_julia(var_name):
    path = os.getenv(var_name)
    if path is None:
        return None, None
    if not os.path.isfile(path):
        raise FileNotFoundError(f"Executable {var_name} = {path} does not exist.")
    version = julia_version.julia_version(path)
    return version, path


def _find_version(version_spec=None, check_exe=False, strict=False, env_var=None, locations=None):
    if version_spec is None:
        version_spec = julia_semver.semver_spec("^1")
    elif isinstance(version_spec, str):
        version_spec = julia_semver.semver_spec(version_spec)
    if locations is None:
        locations = _default_locations()
    if locations['env']:
        if env_var is None:
            env_var = "JULIA"
            env_version, env_path = _env_var_julia(env_var)
            if (env_version is not None and
                julia_semver.match(version_spec, env_version, strict=strict) and
                _check_path(env_path, check_exe)
                ):
                return (env_version, env_path)
    jlists = _collect_paths(locations=locations)
    paths = set(itertools.chain.from_iterable(jlists))
    maxv = julia_semver.version("0.0.0")
    best = (None, None)
    for vers, path in paths:
        if (vers > maxv and julia_semver.match(version_spec, vers, strict=strict)
            and _check_path(path, check_exe)):
            maxv = vers
            best = (vers, path)
    return best


def find(version_spec=None, check_exe=False, find_all=False, strict=False, env_var=None):
    """
    Search for and return the path to a Julia executable.

    Calling `find()` will use reasonable defaults.

    Parameters:
    env_var : The environment variable to check for a julia path.
        If this variable is set and the exectuable satisfies `version_spec`, then it will be
        preferred to all other paths. Default: "JULIA".
    version_spec : A Julia version specification as a str or object. The returned executable
        must satisfy this specification. Default: "^1".
    strict : If `True` then prerelease (development) versions will be excluded.
    check_exe : If `True` then check that the path is a Julia by querying it for the version.
        Note that this has already been done for most Julias found when the version was extracted.
    find_all : If `False` skip the locations that are slower to search. If no other exectuables
        are found, the slower locations may be searched anyway. The only slow location is the
        jill-installed location.
    """
    locations = _default_locations()
    if find_all is False: # Finding jill-installed julia versions is a bit slow
        locations['jill'] = False
    _, path = _find_version(
        version_spec=version_spec, check_exe=check_exe, strict=strict,
        env_var=env_var, locations=locations
    )
    if path is not None:
        return path
    if find_all is not True: # Try other places if no versions were acceptable
        locations = _default_locations(default=False)
        locations['jill'] = True
        _, path = _find_version(
            version_spec=version_spec, check_exe=check_exe, strict=strict, locations=locations
        )
    return path


def find_or_install(version_spec=None, check_exe=False, find_all=False, strict=False,
                    answer_yes=False, post_question_hook=None,
                    env_var=None): # , install_root=None):
    """
    Search for and return the path to a Julia executable or install one if none is found.

    Calling `find_or_install()` will use reasonable defaults.

    This function takes all the same parameters as does `find` as well as the following.
    Parameters:
    answer_yes - if `True`, then ask no questions, assume answers are "yes".
    post_question_hook -  a function to run if and after the consumer is asked whether
        to install Julia. This can be used to ask and record more questions rather
        than waiting till after the download. Default: None
    """
    path = find(
        version_spec=version_spec, check_exe=check_exe, find_all=find_all, strict=strict, env_var=env_var
    )
    if path is not None:
        return path
    util.log(f"No Julia version satisfying spec '{str(version_spec)}' found.")
    install_julia.prompt_and_install(
        answer_yes=answer_yes, post_question_hook=post_question_hook, version_spec=version_spec)
    # find_all=True !!
    path = find(version_spec=version_spec, check_exe=check_exe, find_all=True, strict=strict)
    return path



# This file should only be used until these features are available in jill.py
import os
import getpass
import platform

from .julia_version import to_version_path_list

# import jill
### `install_from_jill` below is the only function that requires jill.
### So, we import jill in the function rather than at the top
### jill.py takes 50-60 ms to load, much longer to load than anything else.

### Code taken from jill.py source
### This is all that is necessary to examine installation locations.
### If we need to install, we load the heavier jill dependency.


def _current_system():
    rst = platform.system()
    if rst.lower() == "linux":
        return "linux"
    if rst.lower() == "darwin":
        return "mac"
    if rst.lower() == "freebsd":
        return "freebsd"
    if rst.lower() == "windows":
        return "winnt"
    raise ValueError(f"Unsupported system {rst}")


def _default_install_dir():
    _dir = os.environ.get("JILL_INSTALL_DIR", None)
    if _dir:
        return os.path.expanduser(_dir)

    system = _current_system()
    if system == "mac":
        return "/Applications"
    if system in ["linux", "freebsd"]:
        if getpass.getuser() == "root":
            return "/opt/julias"
        return os.path.expanduser("~/packages/julias")
    if system == "winnt":
        return os.path.expanduser(r"~\AppData\Local\julias")
    raise ValueError(f"Unsupported system: {system}")

### End code taken from jill.py source

# There is a PR to put the following in jill.py itself. I just have to finish it.
def _get_relative_bin_path():
    system = _current_system()
    if system == "winnt":
        return os.path.join("bin", "julia.exe")
    if system == "mac":
        return os.path.join("Contents", "Resources", "julia", "bin", "julia")
    if system in ["linux", "freebsd"]:
        return os.path.join("bin", "julia")
    raise ValueError(f"Unsupported system {system}")


def _julia_bin_paths(install_root_dir=None):
    if install_root_dir is None:
        install_root_dir = _default_install_dir()
    if not os.path.isdir(install_root_dir):
        return None
    rel_bin_path = _get_relative_bin_path()
    return [os.path.join(install_root_dir, subdir, rel_bin_path)
            for subdir in os.listdir(install_root_dir) if subdir.startswith("julia")]


def version_path_list(install_root_dir=None):
    """
    Return a list of two-tuples of versions and paths of Julias installed by jill.py.
    """
    system = _current_system()
    if install_root_dir is None and system in ["linux", "freebsd"]:
        root = os.path.join("/", "opt", "julias")
        paths1 = _julia_bin_paths(install_root_dir=root)
        root = os.path.expanduser(os.path.join("~", "packages", "julias"))
        paths2 = _julia_bin_paths(install_root_dir=root)
        paths1 = paths1 if paths1 else []
        paths2 = paths2 if paths2 else []
        paths = paths1 + paths2
    else:
        paths = _julia_bin_paths(install_root_dir=install_root_dir)
    return to_version_path_list(paths)


def install_from_jill(answer_yes=False, version=None):
    """
    Use jill.py to install Julia.

    Parameters:
    answer_yes : If `True` ask no questions, assuming answers are "yes".
    version str : The Julia version to install. If `None`, then the latest
       stable release is installed. Default: None
    """
    import jill.install
    result = jill.install.install_julia(confirm=answer_yes, version=version)
    print(f"Install from jill returned '{result}'.") # Return val is complicted, not documented.
    return result
